// 23- (STL )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<assert.h>
#include<list>

#define NODE_SIZE 3

inline int min(int a,int b) {return a<b?a:b;}


//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	node_st *next;								//!<  ,    
	node_st *child;								//!<  
	void Clear();								//!<  
	void Dump(int depth);						//!<  
};

//!  
struct item_st{
	int h;										//!<  
	node_st *node;								//!<  
	item_st();									
	item_st(int v);								
	item_st(int h,node_st*);
	bool Add(node_st*);							//!<      
	int  GetCount();							//!<   
	bool GetMin(int&);							//!<  
};

//!  
class _23heap{
	std::list<item_st>list;						//!<   
	void Insert(item_st*);						//!<     
public:
	typedef std::list<item_st>::iterator iterator;	//!<  
	~_23heap();
	int  Pop();									//!<   
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	void Dump();								//!<  
};


//! \param v  
node_st::node_st(int v)
{
	value=v;
	next=child=0;
}


void node_st::Clear()
{
	if(next) next->Clear();
	if(child) child->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	int n;
	for(n=0; n<depth; n++) putchar(' ');
	printf("[%d]",value);
	if(child || !next) putchar('\n');
	if(child) child->Dump(depth+2);
	if(next)  next->Dump(child?depth:0);
}


//!     (   )
//! \param v    
item_st::item_st(int v)
{
	h=0;
	node=new node_st(v);
}

//! \param h_      
//! \param node_  
item_st::item_st(int h_,node_st *node_)
{
	h=h_;
	node=node_;
}


int item_st::GetCount()
{
	int count=0;
	node_st *temp=node;
	while(temp) count++,temp=temp->next;
	assert(count>=2 && count<=NODE_SIZE);
	return count;
}



//!   ,      
//! \param min  
//! \result true   
bool item_st::GetMin(int &min)
{
	bool res=false;
	node_st *temp=node;
	do{
		if(min>temp->value) min=temp->value,res=true;
		temp=temp->next;
	}while(temp);
	return res;
}



//! \param node_    
//! \result true    
bool item_st::Add(node_st *_node)						
{
	assert(!node->next);
	node->next=_node;
	if(GetCount()<NODE_SIZE) return false;
	h++;
	int min=_node->value; GetMin(min);
	_node=new node_st(min);
	_node->child=node;
	node=_node;
	return true;
}


_23heap::~_23heap()
{
	for(iterator it=list.begin(); it!=list.end(); it++) it->node->Clear();
}


//! \param v  
void _23heap::Insert(int v)
{
	Insert(&item_st(v));
}

//! \param item  
void _23heap::Insert(item_st *item)
{
	item->node->next=0;
	iterator it;
	bool up=true;
	while(up) {
		for(it=list.begin(); it!=list.end(); it++) if(item->h==it->h) break;
		if(it==list.end()) break;
		up=item->Add(it->node);
		list.erase(it);
	}
	list.push_back(*item);
}


bool _23heap::isEmpty()
{
	return list.empty();
}


int _23heap::Pop()
{
	bool unlock;	//     
	iterator it1,it2;
	node_st *node,*child,*next;
	//     
	int min=list.front().node->value+1;
	for(it1=list.begin(); it1!=list.end(); it1++) if(it1->GetMin(min)) it2=it1;
	int h=it2->h;
	child=it2->node;
	list.erase(it2);
	//      ,    
	while(child) {
		node=child; child=0; unlock=true;
		while(node) {
			next=node->next; 
			if(unlock && node->value==min) {	//         
				unlock=false;
				child=node->child;
				delete node;
			} else {							//      ,   
				item_st item(h,node);
				Insert(&item_st(h,node));		//      
			}
			node=next;
		}
		h--;
	}
	return min;
}


void _23heap::Dump()
{
	printf("[dump]\n");
	for(iterator it=list.begin(); it!=list.end(); it++) {
		printf(" -root:%d-\n",it->h);
		it->node->Dump(0);
	}
}



#define SIZE 40

int main()
{
	_23heap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) {
		printf("(%d)",heap.Pop()); 
	}
	putchar('\n');
	printf("[press enter]\n");
	getchar();
	return 0;
}
